try:
    from . import generic as g
except BaseException:
    import generic as g


class TransformTest(g.unittest.TestCase):
    def test_doctest(self):
        """
        Run doctests on transformations, which checks docstrings
        for interactive sessions and then verifies they execute
        correctly.

        This is how the upstream transformations unit tests,
        but it depends on numpy string formatting and is very
        flaky.
        """
        import doctest
        import random

        import trimesh

        # make sure formatting is the same as their docstrings
        g.np.set_printoptions(suppress=True, precision=5)

        # monkey patch import transformations with random for the examples
        trimesh.transformations.random = random

        # search for interactive sessions in docstrings and verify they work
        # they are super unreliable and depend on janky string formatting
        results = doctest.testmod(
            trimesh.transformations, verbose=False, raise_on_error=False
        )

        if results.failed > 0:
            raise ValueError(str(results))
        g.log.debug(str(results))

    def test_downstream(self):
        """
        Run tests on functions that were added by us to the
        original transformations.py
        """
        tr = g.trimesh.transformations

        assert not tr.is_rigid(g.np.ones((4, 4)))

        planar = tr.planar_matrix(offset=[10, -10], theta=0.0)
        assert g.np.allclose(planar[:2, 2], [10, -10])

        planar = tr.planar_matrix(offset=[0, -0], theta=g.np.pi)
        assert g.np.allclose(planar[:2, 2], [0, 0])

        planar = tr.planar_matrix(offset=[0, 0], theta=0.0)
        assert g.np.allclose(planar, g.np.eye(3))

        as_3D = tr.planar_matrix_to_3D(g.np.eye(3))
        assert g.np.allclose(as_3D, g.np.eye(4))

        spherical = tr.spherical_matrix(theta=0.0, phi=0.0)
        assert g.np.allclose(spherical, g.np.eye(4))

        points = g.np.arange(60, dtype=g.np.float64).reshape((-1, 3))
        assert g.np.allclose(tr.transform_points(points, g.np.eye(4)), points)

        points = g.np.arange(60, dtype=g.np.float64).reshape((-1, 2))
        assert g.np.allclose(tr.transform_points(points, g.np.eye(3)), points)

    def test_around(self):
        # check transform_around on 2D points
        points = g.random((100, 2))
        for i, p in enumerate(points):
            offset = g.random(2)
            matrix = g.trimesh.transformations.planar_matrix(
                theta=g.random() + 0.1, offset=offset, point=p
            )

            # apply the matrix
            check = g.trimesh.transform_points(points, matrix)
            compare = g.np.isclose(check, points + offset)
            # the point we rotated around shouldn't move
            assert compare[i].all()
            # all other points should move
            assert compare.all(axis=1).sum() == 1

        # check transform_around on 3D points
        points = g.random((100, 3))
        for i, p in enumerate(points):
            matrix = g.trimesh.transformations.random_rotation_matrix()
            matrix = g.trimesh.transformations.transform_around(matrix, p)

            # apply the matrix
            check = g.trimesh.transform_points(points, matrix)
            compare = g.np.isclose(check, points)
            # the point we rotated around shouldn't move
            assert compare[i].all()
            # all other points should move
            assert compare.all(axis=1).sum() == 1

    def test_rotation(self):
        """
        test
        """
        rotation_matrix = g.trimesh.transformations.rotation_matrix

        R = rotation_matrix(g.np.pi / 2, [0, 0, 1], [1, 0, 0])
        assert g.np.allclose(g.np.dot(R, [0, 0, 0, 1]), [1, -1, 0, 1])

        angle = (g.random() - 0.5) * (2 * g.np.pi)
        direc = g.random(3) - 0.5
        point = g.random(3) - 0.5
        R0 = rotation_matrix(angle, direc, point)
        R1 = rotation_matrix(angle - 2 * g.np.pi, direc, point)
        assert g.trimesh.transformations.is_same_transform(R0, R1)

        R0 = rotation_matrix(angle, direc, point)
        R1 = rotation_matrix(-angle, -direc, point)
        assert g.trimesh.transformations.is_same_transform(R0, R1)

        I = g.np.identity(4, g.np.float64)  # NOQA
        assert g.np.allclose(I, rotation_matrix(g.np.pi * 2, direc))

        assert g.np.allclose(2, g.np.trace(rotation_matrix(g.np.pi / 2, direc, point)))

        # test symbolic
        if g.sp is not None:
            angle = g.sp.Symbol("angle")
            Rs = rotation_matrix(angle, [0, 0, 1], [1, 0, 0])

            R = g.np.array(Rs.subs(angle, g.np.pi / 2.0).evalf()).astype(g.np.float64)

            assert g.np.allclose(g.np.dot(R, [0, 0, 0, 1]), [1, -1, 0, 1])

    def test_tiny(self):
        """
        Test transformations with models containing
        very small triangles.
        """
        for validate in [False, True]:
            m = g.get_mesh("ADIS16480.STL", validate=validate)
            m.apply_scale(0.001)
            m._cache.clear()
            g.np.nonzero(g.np.linalg.norm(m.face_normals, axis=1) < 1e-3)
            m.apply_transform(
                g.trimesh.transformations.rotation_matrix(g.np.pi / 4, [0, 0, 1])
            )

    def test_quat(self):
        """
        Do some simple checks on our quaternion math.
        """
        # shortcuts to long function names
        tf = g.trimesh.transformations
        is_rigid = tf.is_rigid
        multiply = tf.quaternion_multiply
        to_matrix = tf.quaternion_matrix
        from_matrix = tf.quaternion_from_matrix
        random_matrix = tf.random_rotation_matrix
        random_quat = tf.random_quaternion

        # get some arbitrary rotation matrices
        a = tf.rotation_matrix(0.2, g.trimesh.unitize([1, 2, 3]))
        b = tf.rotation_matrix(0.3, g.trimesh.unitize([1, -2, 0]))

        # convert arbitrary rotations to quaternions
        qa = from_matrix(a)
        qb = from_matrix(b)
        # matrix multiply the original matrices
        mm = g.np.dot(a, b)
        # quaternion multiply then convert back to matrix
        qm = to_matrix(multiply(qa, qb))
        # results should be the same
        assert g.np.allclose(mm, qm, atol=1e-5)
        # all random matrices should be rigid transforms
        assert all(is_rigid(T) for T in random_matrix(num=100))
        # random quaternions should all be unit vector
        assert g.np.allclose(
            g.np.linalg.norm(random_quat(num=100), axis=1), 1.0, atol=1e-6
        )

    def test_angle(self):
        assert g.np.isclose(
            g.trimesh.transformations.angle_between_vectors(g.np.ones(3), g.np.ones(3)),
            0.0,
        )

    def test_symbolic_rotation(self):
        # you can pass `sympy.Symbol` to `trimesh.transformation.rotation_matrix`
        try:
            import sympy as sp
        except BaseException:
            return
        tf = g.trimesh.transformations

        a = sp.Symbol("a")
        vector = [1, 1, 1]
        m = tf.rotation_matrix(a, vector)
        for v in [0.0, 1.1, 1.234, g.np.pi]:
            # evaluate the symbolic matrix with a value
            s = g.np.array(m.subs({a: v}).evalf(), dtype=g.np.float64)
            # call rotation matrix with a scalar
            n = tf.rotation_matrix(v, vector)

            # they should be the same matrix
            assert g.np.allclose(s, n)

    def test_symbolic_euler(self):
        # some of the functions have been modified to support `sympy.Symbol`
        # values which is useful for calculating final rotations symbolically
        try:
            import sympy as sp
        except BaseException:
            return

        euler = g.trimesh.transformations.euler_matrix

        ra, rb, rc = sp.symbols("ra rb rc")
        m = euler(ra, rb, rc)
        for rot in g.random((100, 3)):
            # get the euler matrix evaluated from the symbolic matrix
            s = g.np.array(
                m.subs({ra: rot[0], rb: rot[1], rc: rot[2]}).evalf(), dtype=g.np.float64
            )
            # get it from a numeric scalar
            n = euler(*rot)

            # they should be the same matrix
            assert g.np.allclose(s, n)

    def test_symbolic_translate(self):
        # some of the functions have been modified to support `sympy.Symbol`
        # values which is useful for calculating final rotations symbolically
        try:
            import sympy as sp
        except BaseException:
            return

        translate = g.trimesh.transformations.translation_matrix

        x, y, z = sp.symbols("x y z")

        m = translate([x, y, z])

        for T in g.random((100, 3)):
            # get the euler matrix evaluated from the symbolic matrix
            s = g.np.array(
                m.subs({x: T[0], y: T[1], z: T[2]}).evalf(), dtype=g.np.float64
            )
            # get it from a numeric scalar
            n = translate(T)

            # they should be the same matrix
            assert g.np.allclose(s, n)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



"""
https://github.com/mikedh/trimesh
------------------------------------

Trimesh is a pure Python (2.7- 3.3+) library for loading and using triangular
meshes with an emphasis on watertight meshes. The goal of the library is to
provide a fully featured Trimesh object which allows for easy manipulation
and analysis, in the style of the Polygon object in the Shapely library.
"""

# avoid a circular import in trimesh.base
from . import (
    boolean,
    bounds,
    caching,
    collision,
    comparison,
    convex,
    creation,
    curvature,
    decomposition,
    exceptions,
    geometry,
    graph,
    grouping,
    inertia,
    intersections,
    iteration,
    nsphere,
    permutate,
    poses,
    primitives,
    proximity,
    ray,
    registration,
    remesh,
    repair,
    sample,
    smoothing,
    transformations,
    triangles,
    units,
    util,
)
from .base import Trimesh

# general numeric tolerances
from .constants import tol

# loader functions
from .exchange.load import (
    available_formats,
    load,
    load_mesh,
    load_path,
    load_remote,
    load_scene,
)

# geometry objects
from .parent import Geometry
from .points import PointCloud
from .scene.scene import Scene
from .transformations import transform_points

# utility functions
from .util import unitize
from .version import __version__

try:
    # handle vector paths
    from . import path
except BaseException as E:
    # raise a useful error if path hasn't loaded
    path = exceptions.ExceptionWrapper(E)


try:
    from . import voxel
except BaseException as E:
    # requires non-minimal imports
    voxel = exceptions.ExceptionWrapper(E)


__all__ = [
    "Geometry",
    "PointCloud",
    "Scene",
    "Trimesh",
    "__version__",
    "available_formats",
    "boolean",
    "bounds",
    "caching",
    "collision",
    "comparison",
    "convex",
    "creation",
    "curvature",
    "decomposition",
    "exceptions",
    "geometry",
    "graph",
    "grouping",
    "inertia",
    "intersections",
    "iteration",
    "load",
    "load_mesh",
    "load_path",
    "load_remote",
    "load_scene",
    "nsphere",
    "path",
    "permutate",
    "poses",
    "primitives",
    "proximity",
    "ray",
    "registration",
    "remesh",
    "repair",
    "sample",
    "smoothing",
    "tol",
    "transform_points",
    "transformations",
    "triangles",
    "unitize",
    "units",
    "util",
    "voxel",
]



try:
    from . import generic as g
except BaseException:
    import generic as g


class MultiboxTest(g.unittest.TestCase):
    def test_multibox_internal_face_removal(self):
        np = g.np

        # Create a small 2 * 2 * 2 binary voxelgrid
        voxelgrid = np.ones((2, 2, 2), dtype=bool)
        indices = np.argwhere(voxelgrid)
        points = g.trimesh.voxel.ops.indices_to_points(indices)

        # With internal faces (remove_internal_faces=False)
        mesh_w_internal_faces = g.trimesh.voxel.ops.multibox(
            points, pitch=1.0, remove_internal_faces=False
        )

        # Without internal faces (remove_internal_faces=True)
        mesh_wo_internal_faces = g.trimesh.voxel.ops.multibox(
            points, pitch=1.0, remove_internal_faces=True
        )
        self.assertLess(
            len(mesh_wo_internal_faces.faces), len(mesh_w_internal_faces.faces)
        )

        # Check bounding boxes are equal
        g.np.testing.assert_allclose(
            mesh_w_internal_faces.bounds, mesh_wo_internal_faces.bounds
        )

        # A 2 * 2 * 2 voxelgrid constitutes 8 voxels
        # Each voxel has 6 sides, with 2 triangular faces per side, totaling 6 * 2 = 12 triangular faces
        # 8 voxels * 12 triangular faces = 96 triangluar faces (when all the internal faces kept)
        self.assertEqual(len(mesh_w_internal_faces.faces), 96)

        # A 2 * 2 * 2 block has 6 exposed sides (i.e., top, bottom, front, back, left, right)
        # Each exposed side has 4 sub-sides corresponding to the 4 voxels, with 2 triangular faces per sub-side, totaling 4 * 2 = 8 triangular faces
        # 6 exposed sides * 8 triangular face = 48 triangular faces (when all the internal faces removed)
        self.assertEqual(len(mesh_wo_internal_faces.faces), 48)


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()



try:
    from . import generic as g
except BaseException:
    import generic as g


def test_pointcloud():
    """
    Test PointCloud object
    """
    shape = (100, 3)
    # random points
    points = g.random(shape)
    # make sure randomness never gives duplicates by offsetting
    points += g.np.arange(shape[0]).reshape((-1, 1))

    # make some duplicate vertices
    points[:10] = points[0]

    # create a pointcloud object
    cloud = g.trimesh.points.PointCloud(points)

    initial_hash = hash(cloud)

    assert cloud.convex_hull.volume > 0.0

    # check shapes of data
    assert cloud.vertices.shape == shape
    assert cloud.shape == shape
    assert cloud.extents.shape == (3,)
    assert cloud.bounds.shape == (2, 3)

    assert hash(cloud) == initial_hash

    # autogenerate the default weights
    assert cloud.weights.shape == (shape[0],)

    # set some random colors
    cloud.colors = g.random((shape[0], 4))
    # remove the duplicates we created
    cloud.merge_vertices()

    # new shape post- merge
    new_shape = (shape[0] - 9, shape[1])

    # make sure vertices and colors are new shape
    assert cloud.vertices.shape == new_shape
    assert len(cloud.colors) == new_shape[0]
    assert hash(cloud) != initial_hash

    # AABB volume should be same as points
    assert g.np.isclose(cloud.bounding_box.volume, g.np.prod(g.np.ptp(points, axis=0)))

    # will populate all bounding primitives
    assert cloud.bounding_primitive.volume > 0.0
    # ... except AABB (it uses OBB)
    assert cloud.bounding_box.volume > 0.0

    # check getitem and setitem
    cloud[0] = [10, 10, 10]
    assert g.np.allclose(cloud[0], [10, 10, 10])
    # cloud should have copied
    assert not g.np.allclose(points[0], [10, 10, 10])

    # check to see if copy works
    assert g.np.allclose(cloud.vertices, cloud.copy().vertices)
    assert hash(cloud.visual) == hash(cloud.copy().visual)


def test_empty():
    p = g.trimesh.PointCloud(None)
    assert p.is_empty
    assert p.copy().is_empty

    p.vertices = [[0, 1, 2]]
    assert not p.is_empty


def test_vertex_only():
    """
    Test to make sure we can instantiate a mesh with just
    vertices and no faces for some unknowable reason
    """

    v = g.random((1000, 3))
    v[g.np.floor(g.random(90) * len(v)).astype(int)] = v[0]

    mesh = g.trimesh.Trimesh(v)

    assert len(mesh.vertices) < 950
    assert len(mesh.vertices) > 900


def test_plane():
    # make sure plane fitting works for 2D points in space
    for _i in range(10):
        # create a random rotation
        matrix = g.trimesh.transformations.random_rotation_matrix()
        # create some random points in spacd
        p = g.random((1000, 3))
        # make them all lie on the XY plane so we know
        # the correct normal to check against
        p[:, 2] = 0
        # transform them into random frame
        p = g.trimesh.transform_points(p, matrix)
        # we made the Z values zero before transforming
        # so the true normal should be Z then rotated
        truth = g.trimesh.transform_points([[0, 0, 1]], matrix, translate=False)[0]
        # run the plane fit
        _C, N = g.trimesh.points.plane_fit(p)
        # sign of normal is arbitrary on fit so check both
        assert g.np.allclose(truth, N) or g.np.allclose(truth, -N)
    # make sure plane fit works with multiple point sets at once
    nb_points_sets = 20
    for _i in range(10):
        # create a random rotation
        matrices = [
            g.trimesh.transformations.random_rotation_matrix()
            for _ in range(nb_points_sets)
        ]
        # create some random points in spacd
        p = g.random((nb_points_sets, 1000, 3))
        # make them all lie on the XY plane so we know
        # the correct normal to check against
        p[..., 2] = 0
        # transform them into random frame
        for j, matrix in enumerate(matrices):
            p[j, ...] = g.trimesh.transform_points(p[j, ...], matrix)
        # p = g.trimesh.transform_points(p, matrix)
        # we made the Z values zero before transforming
        # so the true normal should be Z then rotated
        truths = g.np.zeros((len(p), 3))
        for j, matrix in enumerate(matrices):
            truths[j, :] = g.trimesh.transform_points(
                [[0, 0, 1]], matrix, translate=False
            )[0]
        # run the plane fit
        _C, N = g.trimesh.points.plane_fit(p)

        # sign of normal is arbitrary on fit so check both
        cosines = g.np.einsum("ij,ij->i", N, truths)
        assert g.np.allclose(g.np.abs(cosines), g.np.ones_like(cosines))


def test_kmeans(cluster_count=5, points_per_cluster=100):
    """
    Test K-means clustering
    """
    clustered = []

    # get random-ish points in a -0.5:+0.5 interval
    points = g.random((points_per_cluster * cluster_count, 3)) - 0.5
    for index, group in enumerate(g.np.array_split(points, cluster_count)):
        # move the cluster along the XYZ vector
        clustered.append((group / 1000) + (index * 100))
    clustered = g.np.vstack(clustered)

    # run k- means clustering on our nicely separated data
    centroids, klabel = g.trimesh.points.k_means(points=clustered, k=cluster_count)

    # reshape to make sure all groups have the same index
    variance = g.np.ptp(klabel.reshape((cluster_count, points_per_cluster)), axis=1)

    assert len(centroids) == cluster_count
    assert (variance == 0).all()


def test_tsp():
    """
    Test our solution for visiting every point in order.
    """
    for dimension in [2, 3]:
        for count in [2, 10, 100]:
            for _i in range(10):
                points = g.random((count, dimension))

                # find a path that visits every point quickly
                idx, dist = g.trimesh.points.tsp(points, start=0)

                # indexes should visit every point exactly once
                assert set(idx) == set(range(len(points)))
                assert len(idx) == len(points)
                assert len(dist) == len(points) - 1

                # shouldn't be any negative indexes
                assert (idx >= 0).all()

                # make sure distances returned are correct
                dist_check = g.np.linalg.norm(g.np.diff(points[idx], axis=0), axis=1)
                assert g.np.allclose(dist_check, dist)


def test_xyz():
    """
    Test XYZ file loading
    """
    # test a small file from cloudcompare
    p = g.get_mesh("points_cloudcompare.xyz")
    assert p.vertices.shape == (101, 3)
    assert p.colors.shape == (101, 4)

    # test a small file from agisoft
    p = g.get_mesh("points_agisoft.xyz")
    assert p.vertices.shape == (100, 3)
    assert p.colors.shape == (100, 4)

    # test exports
    e = p.export(file_type="xyz")
    p = g.trimesh.load(g.trimesh.util.wrap_as_stream(e), file_type="xyz")
    assert p.vertices.shape == (100, 3)
    assert p.colors.shape == (100, 4)


def test_obb():
    p = g.get_mesh("points_agisoft.xyz")
    original = p.bounds.copy()
    matrix = p.apply_obb()
    assert matrix.shape == (4, 4)
    assert not g.np.allclose(p.bounds, original)


def test_ply():
    p = g.get_mesh("points_agisoft.xyz")
    assert isinstance(p, g.trimesh.PointCloud)
    assert len(p.vertices) > 0

    # initial color CRC
    initial = hash(p.visual)
    # set to random colors
    p.colors = g.random((len(p.vertices), 4))
    # visual CRC should have changed
    assert hash(p.visual) != initial

    # test exporting a pointcloud to a PLY file
    r = g.roundtrip(p.export(file_type="ply"), file_type="ply")
    assert r.vertices.shape == p.vertices.shape
    # make sure colors survived the round trip
    assert g.np.allclose(r.colors, p.colors)


def test_glb():
    p = g.get_mesh("points_agisoft.xyz")
    assert isinstance(p, g.trimesh.PointCloud)
    assert len(p.vertices) > 0
    # test exporting a pointcloud to a GLTF
    # TODO : WE SHOULD IMPLEMENT THE IMPORTER TOO
    r = p.export(file_type="gltf")
    assert len(g.json.loads(r["model.gltf"].decode("utf-8"))["meshes"]) == 1


def test_remove_close():
    # create 100 unique points
    p = g.np.arange(300).reshape((100, 3))
    # should return the original 100 points
    culled, mask = g.trimesh.points.remove_close(g.np.vstack((p, p)), radius=0.1)
    assert culled.shape == (100, 3)
    assert mask.shape == (200,)


def test_add_operator():
    points_1 = g.random((10, 3))
    points_2 = g.random((20, 3))
    colors_1 = [[123, 123, 123, 255]] * len(points_1)
    colors_2 = [[255, 0, 123, 255]] * len(points_2)

    # Test: Both cloud's colors are preserved
    cloud_1 = g.trimesh.points.PointCloud(points_1, colors=colors_1)
    cloud_2 = g.trimesh.points.PointCloud(points_2, colors=colors_2)

    cloud_sum = cloud_1 + cloud_2
    assert g.np.allclose(cloud_sum.colors, g.np.vstack((cloud_1.colors, cloud_2.colors)))

    # Next test: Only second cloud has colors
    cloud_1 = g.trimesh.points.PointCloud(points_1)
    cloud_2 = g.trimesh.points.PointCloud(points_2, colors=colors_2)

    cloud_sum = cloud_1 + cloud_2
    assert g.np.allclose(cloud_sum.colors[len(cloud_1.vertices) :], cloud_2.colors)

    # Next test: Only first cloud has colors
    cloud_1 = g.trimesh.points.PointCloud(points_1, colors=colors_1)
    cloud_2 = g.trimesh.points.PointCloud(points_2)

    cloud_sum = cloud_1 + cloud_2
    assert g.np.allclose(cloud_sum.colors[: len(cloud_1.vertices)], cloud_1.colors)


def test_radial_sort():
    theta = g.np.linspace(0.0, g.np.pi * 2.0, 1000)
    points = g.np.column_stack((g.np.cos(theta), g.np.sin(theta), g.np.zeros(len(theta))))
    points *= g.random(len(theta)).reshape((-1, 1))

    # apply a random order to the points
    order = g.np.random.permutation(g.np.arange(len(points)))

    # get the sorted version of these points
    # when we pass them the randomly ordered points
    sort = g.trimesh.points.radial_sort(points[order], origin=[0, 0, 0], normal=[0, 0, 1])
    # should have re-established original order
    assert g.np.allclose(points, sort)



import unittest

import numpy as np

import trimesh


class PBRTest(unittest.TestCase):
    def test_storage(self):
        p = trimesh.visual.material.PBRMaterial()
        assert isinstance(p, trimesh.visual.material.PBRMaterial)
        a = [hash(p)]

        try:
            # should raise a ValueError as a float
            # is non-convertable to RGBA colors
            p.baseColorFactor = 1.0
            raise BaseException("should have disallowed!")
        except ValueError:
            pass

        assert p.baseColorFactor is None

        # apply a basecolorfactor
        original = [255, 255, 100]
        p.baseColorFactor = original
        assert len(p.baseColorFactor) == 4
        assert np.allclose(p.baseColorFactor, np.append(original, 255))

        a.append(hash(p))
        # hash should have changed when we edited color
        assert a[-1] != a[-2]
        # hashing should be deterministic
        assert hash(p) == a[-1]

        assert p.alphaMode is None
        try:
            # only allowed to be 3 things not including a sandwich
            p.alphaMode = "sandwich"
            raise BaseException("shouldn't have passed")
        except ValueError:
            pass
        p.alphaMode = "OPAQUE"
        assert p.alphaMode == "OPAQUE"

        a.append(hash(p))
        # hash should have changed when we edited alphaMode
        assert a[-1] != a[-2]

        assert p.emissiveFactor is None
        try:
            # only allowed to be 3 things not including a sandwich
            p.emissiveFactor = "sandwitch"
            raise BaseException("shouldn't have passed")
        except ValueError:
            pass
        try:
            # only allowed to be (3,) float
            p.emissiveFactor = [1, 2]
            raise BaseException("shouldn't have passed")
        except ValueError:
            pass

        # shouldn't have changed yet
        assert hash(p) == a[-1]
        assert p.emissiveFactor is None

        # set emissive to a valid value
        p.emissiveFactor = [1, 1, 0.5]
        a.append(hash(p))
        assert a[-1] != a[-2]
        assert np.allclose(p.emissiveFactor, [1, 1, 0.5])

        try:
            # only allowed to be float
            p.roughnessFactor = "hi"
            raise BaseException("shouldn't have passed")
        except ValueError:
            pass
        # shouldn't have changed yet
        assert hash(p) == a[-1]
        assert p.roughnessFactor is None
        # set to valid value
        original = 0.023
        p.roughnessFactor = original
        assert np.isclose(original, p.roughnessFactor)
        a.append(hash(p))
        assert a[-1] != a[-2]

        try:
            # only allowed to be float
            p.metallicFactor = "hi"
            raise BaseException("shouldn't have passed")
        except ValueError:
            pass
        # shouldn't have changed yet
        assert hash(p) == a[-1]
        assert p.metallicFactor is None
        # set to valid value
        original = 0.9113
        p.metallicFactor = original
        assert np.isclose(original, p.metallicFactor)
        a.append(hash(p))
        assert a[-1] != a[-2]

        # double check all of our hashes changed
        assert (np.abs(np.diff(a)) > 0).all()


if __name__ == "__main__":
    trimesh.util.attach_to_log()
    unittest.main()



import numpy as np

from .typed import ArrayLike, Optional

try:
    from scipy.sparse import coo_matrix, eye
    from scipy.sparse.linalg import spsolve
except ImportError as E:
    from .exceptions import ExceptionWrapper

    wrapper = ExceptionWrapper(E)
    eye, spsolve, coo_matrix = wrapper, wrapper, wrapper

from . import graph, triangles
from .base import Trimesh
from .geometry import index_sparse
from .triangles import mass_properties
from .util import unitize


def filter_laplacian(
    mesh,
    lamb=0.5,
    iterations=10,
    implicit_time_integration=False,
    volume_constraint=True,
    laplacian_operator=None,
):
    """
    Smooth a mesh in-place using laplacian smoothing.
    Articles
    1 - "Improved Laplacian Smoothing of Noisy Surface Meshes"
       J. Vollmer, R. Mencl, and H. Muller
    2 - "Implicit Fairing of Irregular Meshes using Diffusion
       and Curvature Flow". M. Desbrun,  M. Meyer,
       P. Schroder, A.H.B. Caltech
    Parameters
    ------------
    mesh : trimesh.Trimesh
    Mesh to be smoothed in place
    lamb : float
    Diffusion speed constant
    If   0.0, no diffusion
    If > 0.0, diffusion occurs
    implicit_time_integration: boolean
    if False: explicit time integration
        -lamb <= 1.0 - Stability Limit (Article 1)
    if True: implicit time integration
        -lamb no limit (Article 2)
    iterations : int
    Number of passes to run filter
    laplacian_operator : None or scipy.sparse.coo.coo_matrix
    Sparse matrix laplacian operator
    Will be autogenerated if None
    """

    # if the laplacian operator was not passed create it here
    if laplacian_operator is None:
        laplacian_operator = laplacian_calculation(mesh)

    # save initial volume
    if volume_constraint:
        vol_ini = mesh.volume

    # get mesh vertices and faces as vanilla numpy array
    vertices = mesh.vertices.copy().view(np.ndarray)
    faces = mesh.faces.copy().view(np.ndarray)

    # Set matrix for linear system of equations
    if implicit_time_integration:
        dlap = laplacian_operator.shape[0]
        AA = eye(dlap) + lamb * (eye(dlap) - laplacian_operator)

    # Number of passes
    for _index in range(iterations):
        # Classic Explicit Time Integration - Article 1
        if not implicit_time_integration:
            dot = laplacian_operator.dot(vertices) - vertices
            vertices += lamb * dot

        # Implicit Time Integration - Article 2
        else:
            vertices = spsolve(AA, vertices)

        # volume constraint
        if volume_constraint:
            # find the volume with new vertex positions
            vol_new = triangles.mass_properties(vertices[faces], skip_inertia=True)[
                "volume"
            ]
            # scale by volume ratio
            vertices *= (vol_ini / vol_new) ** (1.0 / 3.0)

    # assign modified vertices back to mesh
    mesh.vertices = vertices
    return mesh


def filter_humphrey(mesh, alpha=0.1, beta=0.5, iterations=10, laplacian_operator=None):
    """
    Smooth a mesh in-place using laplacian smoothing
    and Humphrey filtering.
    Articles
    "Improved Laplacian Smoothing of Noisy Surface Meshes"
    J. Vollmer, R. Mencl, and H. Muller
    Parameters
    ------------
    mesh : trimesh.Trimesh
      Mesh to be smoothed in place
    alpha : float
      Controls shrinkage, range is 0.0 - 1.0
      If 0.0, not considered
      If 1.0, no smoothing
    beta : float
      Controls how aggressive smoothing is
      If 0.0, no smoothing
      If 1.0, full aggressiveness
    iterations : int
      Number of passes to run filter
    laplacian_operator : None or scipy.sparse.coo.coo_matrix
      Sparse matrix laplacian operator
      Will be autogenerated if None
    """
    # if the laplacian operator was not passed create it here
    if laplacian_operator is None:
        laplacian_operator = laplacian_calculation(mesh)

    # get mesh vertices as vanilla numpy array
    vertices = mesh.vertices.copy().view(np.ndarray)
    # save original unmodified vertices
    original = vertices.copy()

    # run through iterations of filter
    for _index in range(iterations):
        vert_q = vertices.copy()
        vertices = laplacian_operator.dot(vertices)
        vert_b = vertices - (alpha * original + (1.0 - alpha) * vert_q)
        vertices -= beta * vert_b + (1.0 - beta) * laplacian_operator.dot(vert_b)

    # assign modified vertices back to mesh
    mesh.vertices = vertices
    return mesh


def filter_taubin(mesh, lamb=0.5, nu=0.5, iterations=10, laplacian_operator=None):
    """
    Smooth a mesh in-place using laplacian smoothing
    and taubin filtering.
    Articles
    "Improved Laplacian Smoothing of Noisy Surface Meshes"
    J. Vollmer, R. Mencl, and H. Muller
    Parameters
    ------------
    mesh : trimesh.Trimesh
      Mesh to be smoothed in place.
    lamb : float
      Controls shrinkage, range is 0.0 - 1.0
    nu : float
      Controls dilation, range is 0.0 - 1.0
      Nu shall be between 0.0 < 1.0/lambda - 1.0/nu < 0.1
    iterations : int
      Number of passes to run the filter
    laplacian_operator : None or scipy.sparse.coo.coo_matrix
      Sparse matrix laplacian operator
      Will be autogenerated if None
    """
    # if the laplacian operator was not passed create it here
    if laplacian_operator is None:
        laplacian_operator = laplacian_calculation(mesh)

    # get mesh vertices as vanilla numpy array
    vertices = mesh.vertices.copy().view(np.ndarray)

    # run through multiple passes of the filter
    for index in range(iterations):
        # do a sparse dot product on the vertices
        dot = laplacian_operator.dot(vertices) - vertices
        # alternate shrinkage and dilation
        if index % 2 == 0:
            vertices += lamb * dot
        else:
            vertices -= nu * dot

    # assign updated vertices back to mesh
    mesh.vertices = vertices
    return mesh


def filter_mut_dif_laplacian(
    mesh, lamb=0.5, iterations=10, volume_constraint=True, laplacian_operator=None
):
    """
    Smooth a mesh in-place using laplacian smoothing using a
    mutable diffusion laplacian.

    Articles
      Barroqueiro, B., Andrade-Campos, A., Dias-de-Oliveira,
      J., and Valente, R. (January 21, 2021).
      "Bridging between topology optimization and additive
       manufacturing via Laplacian smoothing." ASME. J. Mech. Des.


    Parameters
    ------------
    mesh : trimesh.Trimesh
      Mesh to be smoothed in place
    lamb : float
      Diffusion speed constant
      If   0.0, no diffusion
      If > 0.0, diffusion occurs
    iterations : int
      Number of passes to run filter
    laplacian_operator : None or scipy.sparse.coo.coo_matrix
      Sparse matrix laplacian operator
      Will be autogenerated if None
    """

    # if the laplacian operator was not passed create it here
    if laplacian_operator is None:
        laplacian_operator = laplacian_calculation(mesh)

    # Set volume constraint
    if volume_constraint:
        v_ini = mesh.volume

    # get mesh vertices as vanilla numpy array
    vertices = mesh.vertices.copy().view(np.ndarray)
    faces = mesh.faces.copy().view(np.ndarray)
    eps = 0.01 * (np.max(mesh.area_faces) ** 0.5)

    # Number of passes
    for _index in range(iterations):
        # Mutable diffusion
        normals = get_vertices_normals(mesh)
        qi = laplacian_operator.dot(vertices)
        pi_qi = vertices - qi

        adil = np.abs((normals * pi_qi).dot(np.ones((3, 1))))
        adil = 1.0 / np.maximum(1e-12, adil)
        lamber = np.maximum(0.2 * lamb, np.minimum(1.0, lamb * adil / np.mean(adil)))

        # Filter
        dot = laplacian_operator.dot(vertices)
        vertices += lamber * (dot - vertices)

        # Volume constraint
        if volume_constraint:
            vol = mass_properties(vertices[faces], skip_inertia=True)["volume"]
            if _index == 0:
                slope = dilate_slope(vertices, faces, normals, vol, eps)
            vertices += normals * slope * (v_ini - vol)

    # assign modified vertices back to mesh
    mesh.vertices = vertices

    return mesh


def laplacian_calculation(
    mesh: Trimesh,
    equal_weight: bool = True,
    pinned_vertices: Optional[ArrayLike] = None,
):
    """
    Calculate a sparse matrix for laplacian operations.

    Note that setting equal_weight to False significantly hampers performance.

    Parameters
    -------------
    mesh : trimesh.Trimesh
      Input geometry
    equal_weight : bool
      If True, all neighbors will be considered equally
      If False, all neighbors will be weighted by inverse distance
    pinned_vertices : None or list of ints
      If None, no vertices are pinned
      If list, vertices will be pinned, such that they will not be moved
    Returns
    ----------
    laplacian : scipy.sparse.coo.coo_matrix
      Laplacian operator
    """
    if equal_weight:
        laplacian = graph.edges_to_coo(mesh.edges)

        if pinned_vertices is not None:
            # Set pinned vertices to only have themselves as neighbours
            laplacian.data[np.isin(laplacian.row, pinned_vertices)] = False
            laplacian.row = np.concatenate((laplacian.row, pinned_vertices))
            laplacian.col = np.concatenate((laplacian.col, pinned_vertices))
            laplacian.data = np.concatenate(
                (laplacian.data, np.ones(len(pinned_vertices), dtype=bool))
            )

        # using this instead of the line below
        # laplacian = laplacian / laplacian.sum(axis=1)
        # because on Python 3.8 that returns a `numpy.matrix`
        # value instead of a `sparse.coo_matrix`
        laplacian = laplacian.multiply(1.0 / laplacian.sum(axis=1))

    else:
        # get the vertex neighbors from the cache
        neighbors = mesh.vertex_neighbors

        # if a node is pinned, it will average his coordinates by himself
        # in practice it will not move
        if pinned_vertices is not None:
            for i in pinned_vertices:
                neighbors[i] = [i]

        # avoid hitting crc checks in loops
        vertices = mesh.vertices.view(np.ndarray)

        # stack neighbors to 1D arrays
        col = np.concatenate(neighbors)
        row = np.concatenate([[i] * len(n) for i, n in enumerate(neighbors)])

        # umbrella weights, distance-weighted
        # use dot product of ones to replace array.sum(axis=1)
        ones = np.ones(3)
        # the distance from verticesex to neighbors
        norms = [
            1.0
            / np.maximum(1e-6, np.sqrt(np.dot((vertices[i] - vertices[n]) ** 2, ones)))
            for i, n in enumerate(neighbors)
        ]
        # normalize group and stack into single array
        data = np.concatenate([i / i.sum() for i in norms])

        # create the sparse matrix
        laplacian = coo_matrix((data, (row, col)), shape=[len(vertices)] * 2)

    return laplacian


def get_vertices_normals(mesh):
    """
    Compute Vertex normals using equal weighting of neighbors faces.
    Parameters
      -------------
      mesh : trimesh.Trimesh
        Input geometry
      Returns
      ----------
      vertices_normals: array
        Vertices normals
    """

    # get mesh vertices and faces
    vertices = mesh.vertices
    faces = mesh.faces

    # get face normals
    face_normals = mesh.face_normals

    # Compute Vert normals
    vert_normals = index_sparse(len(vertices), faces).dot(face_normals)

    return unitize(vert_normals)


def dilate_slope(vertices, faces, normals, v, eps):
    """
    Get the derivate of dilation scalar by the volume variation by finite differences
    Thus, Vertices += vertex_normals*dilate_slope*(Initial_Volume - Srinked_Volume)
    Parameters
      -------------
      mesh : trimesh.Trimesh
        Input geometry
      vertices: mesh.vertices
      faces: mesh.faces
      normals: array
        vertices normals
      Returns
      ----------
      dilate_slope: float
        derivative
    """

    # finite difference derivative
    vertices2 = vertices + normals * eps
    v2 = mass_properties(vertices2[faces], skip_inertia=True)["volume"]

    return (eps) / (v2 - v)



try:
    import generic as g
except BaseException:
    from . import generic as g


class STEPTests(g.unittest.TestCase):
    def test_basic(self):
        try:
            import cascadio  # noqa
        except BaseException:
            g.log.error("failed to get cascadio!", exc_info=True)
            return

        s = g.get_mesh("box_sides.STEP")
        assert len(s.geometry) == 10

        b = g.get_mesh("featuretype.STEP")
        assert len(b.geometry) == 1
        m = next(iter(b.geometry.values()))
        m.merge_vertices(merge_tex=True, merge_norm=True)
        assert m.is_watertight


if __name__ == "__main__":
    g.trimesh.util.attach_to_log()
    g.unittest.main()
